#!/usr/bin/env python


from __future__ import print_function, division

import datetime
import argparse
import numpy as np
import os.path
import pandas as pd

import pypfilt
import epifx

from observer import VDHHS
from llestimators2 import FitLikelihood, ForecastLikelihood
from llestimators2 import PredictiveCIs


def get_params(start, end, args):
    """Return simulation parameters.

    Note the use of SeasonalSEIR imported from model.py.

    :param start: Simulation start date.
    :param end: Simulation stop date.
    :param args: Dictionary containing information from the command-line
        arguments.
    """
    p_exp = 10 ** -2
    px_count = args['px-count']
    px_scale = p_exp * px_count
    params = epifx.default_params(p_exp, px_scale)


    params['obs']['sim_start'] = start
    params['obs']['sim_end'] = end
    epifx.init_prng(params, 1)

    # Set a fixed PRNG seed for re-sampling.
    params['resample']['prng_seed'] = 1

    if args['forcing'] is not None:
        params['epifx']['forcing'] = epifx.daily_forcing(args['forcing'])
        params['prior']['sigma'] = lambda r, size=None: r.uniform(-0.2, 0,
                                                                  size=size)
        params['param_min'][4] = -0.2
        params['param_max'][4] = 0

    if not args['inhomogeneous']:
        # Keep eta fixed at 1 (i.e., enforce homogeneous mixing), otherwise the
        # default behaviour is to assume eta has a uniform prior from 1 to 2,
        # since we are working with a normalised population.
        eta_fixed = 1.0
        params['prior']['eta'] = lambda r, size=None: eta_fixed
        params['param_min'][3] = eta_fixed
        params['param_max'][3] = eta_fixed

    if args['R0_max'] != 2.0:
        params['prior']['R0'] = lambda r, size=None: r.uniform(1.0,
                                                               args['R0_max'],
                                                               size=size)
        params['param_min'][0] = 1 / 3
        params['param_max'][0] = args['R0_max']
    # Toggle the noise on the flow between compartments if a value is given
    # other than the default.
    if args['noise_flow'] != 0.025:
        params['prior']['noise_flow'] = lambda r: np.array(args['noise_flow'])
    # By default regularisation is turned off and the parameters get a random
    # perturbation at each time step. If the user wants to use regularisation
    # this must be activated and the random perturbations turned off.
    if args['regularisation']:
        params['resample']['regularisation'] = True  # Default is False.
        params['prior']['noise_param'] = lambda r: np.array(0)
    return params


def get_streams(params, year, args):
    """Returns a list of observation streams and specifies an observation model
    in params.

    :param params: Simulation parameters.
    :param year: Influenza season under consideration.
    :param pr_notif: The probability a case is observed.
    :param disp: The dispersion parameter.
    """
    streams = [VDHHS.obs(params['obs']['sim_start'], params['obs']['sim_end'])]
    obs_unit = VDHHS.obs_unit
    params['obs'][obs_unit] = dict(log_llhd_fn=VDHHS.log_llhd,
                                   expect_fn=VDHHS.expect)
    # Optimal values from Moss 2016.
    bgs = {2010: 15, 2011: 32, 2012: 28, 2013: 32, 2014: 46, 2015: 60}
    pr_notif_infs = {2010: 0.00025, 2011: 0.002, 2012: 0.001, 2013: 0.002,
                     2014: 0.00275, 2015: 0.01025}
    disps = {y: 100 for y in range(2010, 2016)}
    params['obs'][obs_unit]['obs_bg'] = (bgs[year] if (args['obs_bg'] is None)
                                         else args['obs_bg'])
    params['obs'][obs_unit]['pr_notif_inf'] = (pr_notif_infs[year] if
                                               (args['pr_notif_inf'] is None)
                                               else args['pr_notif_inf'])
    params['obs'][obs_unit]['disp'] = (disps[year] if (args['disp'] is None)
                                       else args['disp'])
    return streams


def get_summary(params, streams, args):
    """Returns a summary (HDF5) object.

    This is where the extra Table objects can be included.
    To avoid eating up the wrong disk space this will write the temporary files
    to the same directory that the output is to be written to. To do this the
    value of `'temp_dir'` in `params` is modified.

    :param params: Simulation parameters.
    :param streams: List of observation streams.
    :param args: Dictionary containin the command line arguments.
    """
    # If no filename is provided give a unique identifier.
    if args['out_file'] == '':
        filename = 'sim-results/{}.hdf5'.format(
            datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S"))
    else:
        filename = args['out_file']
    # We want the temporary files to be stored in the same location as the
    # output files as presumably this has been chosen as a good location to
    # dump a lot of data.
    if '/' in filename:
        params['tmp_dir'] = '/'.join(filename.split('/')[:-1])

    tables = []
    # Allow user to specify resampling prior to likelihood estimation.
    if args['resample_table'] is None:
        tables.append(FitLikelihood())
    else:
        tables.append(FitLikelihood(resample_method=args['resample_table']))

    tables.append(ForecastLikelihood())
    if args['predictive_cis']:
        tables.append(PredictiveCIs(probs=np.uint8([0, 50, 90, 95])))

    summary = epifx.summary.make(params, streams[0], extra_tbls=tables)
    return summary, filename


def get_observation_dates(filename, year, prior_to_peak=8, post_peak=0):
    """Return a list of datetime objects describing when observations where
    made so these can be used as forecasting dates.

    :param filename: The file to read the observations out of.
    :param prior_to_peak: The number of observations prior to the return.
    :param post_peak: The number of observations after the peak to return.
    """
    dtype = [('year', np.int16), ('week', np.int16), ('from', str), ('to', str),
             ('count', np.int16)]
    obs = pd.read_csv(filename, sep=' ', dtype=dtype)
    obs = obs[obs['year'] == year]
    max_count_ix = obs['count'].idxmax()
    ixs = [max_count_ix - prior_to_peak + i for i in range(prior_to_peak + 1)]
    ixs += [max_count_ix + 1 + i for i in range(post_peak)]
    return [datetime.datetime.strptime(obs['to'].loc[i], '%Y-%m-%d') for i in ixs]


def _parse_args():
    """Parse the command-line arguments.

    :returns: A dictionary containing the information taken from the
        command-line arguments.
    """
    p = argparse.ArgumentParser()
    # Information about what to run.
    ip = p.add_argument_group(title='Input')
    ip.add_argument('-d', '--dates', nargs=2, default=None, type=str,
                    required=True, help='Specify the start and end date as '
                                        'strings with format %%Y-%%m-%%d, i.e.,'
                                        ' yyyy-mm-dd.')
    ip.add_argument('-f', '--forcing', default=None, type=str,
                    help='Specify the forcing file. The default behaviour is '
                         'to **not** include forcing.')
    ip.add_argument('--inhomogeneous', action='store_true',
                    help='Turn on inhomogeneous mixing.')
    ip.add_argument('-r', '--resample-table', default='basic', type=str,
                    choices=['basic', 'stratified', 'deterministic'],
                    help='Specify the type of resampling used in the '
                         'loglikelihood estimation.')
    ip.add_argument('-p', '--period', default=1, type=int,
                    help='Specify the integer period of forecasting. Defaults '
                         'to 1 meaning that a forecast is produced after every '
                         'weekly observation.')
    ip.add_argument('--px-count', default=10**3, type=int,
                    help='Specify the number of particles to use in the '
                         'simulation. This defaults to 10^3.', dest='px-count')
    ip.add_argument('--noise-flow', default=0.025, type=float,
                    help='This allows for control of the how much noise there '
                         'is in the flow between compartments. The default '
                         'value is 0.025 which is what is implemented in'
                         'epifx.')
    ip.add_argument('-R', '--R0-max', default=2.0, type=float,
                    help='Set the maximum value of R-naught in the '
                         'simulations.')
    ip.add_argument('-o', '--observation-file', default=None, type=str,
                    help='The path to the file containing the VDHHS '
                         'observations.')
    ip.add_argument('--resample-forecast', action='store_true',
                    help='This flag specifies that the particles should be '
                         'resampled prior to generating any forecasts.')
    ip.add_argument('--forecast-horizon', default=None, type=int,
                    help='Set the number of weeks to estimate the '
                         'log-likelihood of under the forecasting '
                         'distribution.', dest='forecast_horizon')
    ip.add_argument('--forecast-weeks', nargs=2, default=[8, 0], type=int,
                    help='Specify the number of observations to generate '
                         'forecasts on prior to and after the peak week '
                         'respectively.', dest='forecast_weeks')
    ip.add_argument('--predictive-cis', action='store_true',
                    help='This flag specifies that the marginal predictive '
                         'credible intervals should be approximated.')
    # Specification of the observation model.
    obs_mod = p.add_argument_group(title = 'Observation model')
    obs_mod.add_argument('--disp', default=None, type=float,
                         help='Dispersion parameter.')
    obs_mod.add_argument('--obs-bg', default=None, type=float,
                         help='Background rate (notifications per week).')
    obs_mod.add_argument('--pr-notif-inf', default=None, type=float,
                         help='Probability of observing epidemic case.')
    # Particle filter specification.
    pf = p.add_argument_group(title='Particle filter')
    pf.add_argument('--regularisation', action='store_true',
                    help='Turn on regularisation and turn off (default) Weiner '
                         'process on parameters.')
    # Information about what to output.
    op = p.add_argument_group(title='Output')
    op.add_argument('--out-file', default='', type=str,
                    help='Specify the name of the output file.')
    a = vars(p.parse_args())
    a['dates'] = [datetime.datetime.strptime(x, '%Y-%m-%d')
                  for x in a['dates']]

    assert a['dates'][0] < a['dates'][1]
    if a['out_file']:
        assert not os.path.isfile(a['out_file'])
    if a['forcing']:
        assert os.path.isfile(a['forcing'])

    return a


def _check_dates(start, end, streams):
    """Perform a quick check that the start and end date are appropriate."""
    obs_dates = [obs_dict['date'] for obs_dict in streams[0]]
    msg = 'Recommended {} date: {}'
    assert start != (min(obs_dates) - datetime.timedelta(days=6))
    assert end in obs_dates, msg.format('end', max(obs_dates))
    return 0


def main(args):
    start = args['dates'][0]
    end = args['dates'][1]

    params = get_params(start, end, args)
    streams = get_streams(params, start.year, args)
    _check_dates(start, end, streams)
    dates = get_observation_dates(args['observation_file'], start.year,
                                  prior_to_peak=args['forecast_weeks'][0],
                                  post_peak=args['forecast_weeks'][1])
    summary, filename = get_summary(params, streams, args)

    if args['resample_forecast']:
        # Produce forecasts after forcing a resampling of the particles.
        resampledForecast.resampled_forecast(params, start, end, streams, dates,
                                             summary, filename)
    else:
        pypfilt.forecast(params, start, end, streams, dates, summary, filename)
    return 0

if __name__ == "__main__":
    args = _parse_args()
    main(args)
