from __future__ import absolute_import, division, print_function
# from __future__ import unicode_literals

import copy
import datetime
import numpy as np
from scipy.stats import norm

import pypfilt
from pypfilt.summary import dtype_date, Table, Monitor
from pypfilt import stats
from epifx.summary import PeakMonitor


class FitLikelihood(Table):
    """
    Estimate the likelihood of the observed data.

    In this Table stochastic integration is done with an uniformly weighted
    sample so it sufficies to use numpy's mean function rather than the
    pypfilt.stats method. Moreover there does not appear to be any need to sort
    the numbers before adding them as a comparison yielded no difference.
    """
    def __init__(self, name='fit_likelihood', seed=7, resample_method='basic'):
        Table.__init__(self, name)
        self.__rng = np.random.RandomState(seed)
        if resample_method in ['basic', 'stratified', 'determinisitc']:
            self._resample_method = resample_method
        else:
            raise ValueError

    def _copy_params(self, params):
        """Return a copy of the params dictionary with the epifx and pypfilt
        PRNGs replaced by the local PRNG.

        This is necessary so that when we make a call to the update method of
        the model we do not change the state of any random number generators.
        This makes it possible to include this Table without causing any
        changes to the results.
        """
        pc = copy.deepcopy(params)
        # Depending on whether an independent PRNG is used this needs to be
        # stored in a different place.
        if pc['epifx']['independent_prng']:
            pc['epifx']['rnd'] = self.__rng
        else:
            pc['resample']['rnd'] = self.__rng
        pc['size'] = params['hist']['px_count']
        return pc

    def dtype(self, params, obs_list):
        """
        Each row contains:
         - the observation date
         - an estimate of the likelihood
         - an estimate of the standard error (when using basic resampling)
        """
        self.__params_original = params
        self.__params_safe = self._copy_params(params)
        self.__dt = 1.0 / params['steps_per_day']
        self.__sc = params['model'].state_size()
        self.__model = params['model']
        self.__log_llhd = params['log_llhd_fn']  # **LOG**-likelihood function.
        self.__op = params['obs']
        self.__obs_list = obs_list
        self._obs_dates = [o['date'] for o in obs_list]
        self.__obs_periods = set(o['period'] for o in obs_list)
        self.__obs_dict = {d: [] for d in self._obs_dates}
        for o in obs_list:
            self.__obs_dict[o['date']].append(o)

        obs_date = dtype_date('obs_date')
        lhd = ('likelihood', np.float64)
        se = ('standard_error', np.float64)
        if self._resample_method == 'basic':
            return [obs_date, lhd, se]
        else:
            return [obs_date, lhd]

    def _estimate_log_llhd(self, obs_list, hist, date, hist_ix):
        """Estimate the **log**-likelihood of the observations.

        This uses Monte Carlo integration to estimate the log-likelihood of the
        observations. Care must be taken to ensure that the history array
        represents a density with the correct conditioning since this class
        does not make any assumptions about this.
        """
        hist_copy = np.copy(hist)
        prev = hist_copy[hist_ix - 1, ...]
        curr = hist_copy[hist_ix, ...]
        curr *= 0  # erase information from the current observation.
        # re-sample based on the density just prior to the current observation
        # to obtain a random sample from the correct density.
        pseudo_params = dict(resample=dict(method=self._resample_method,
                                           regularisation=False,
                                           rnd=self.__rng))
        pypfilt.resample(pseudo_params, prev)
        # Since resample() only works on one time slice of hist_copy we need to
        # apply the same changes to the earlier part of the history array.
        resampled_ixs = np.copy(prev[:, -1]).astype(int)
        hist_copy[:(hist_ix - 2), ...] = hist_copy[:(hist_ix - 2),
                                                   resampled_ixs, :]
        # integrate to obtain the particle states at the current time. See
        # pypfilt.step() for details.
        prev_sc = prev[..., 0:self.__sc]
        curr_sc = curr[..., 0:self.__sc]
        when = date - datetime.timedelta(days=self.__dt)

        self.__model.update(self.__params_safe, when, self.__dt, None, prev_sc,
                            curr_sc)
        curr[:, -2] = prev[:, -2]
        # since there are potentially multiple observations from different
        # streams we need the periods for each stream.

        def hist_for(period):
            """Return past state vectors in the appropriate order."""
            steps_back = self.__params_original['steps_per_day'] * period
            if hist_ix - steps_back < 0:
                raise ValueError('Attempting to access distant past.')
            return hist_copy[hist_ix - steps_back]

        period_hists = {period: hist_for(period)
                        for period in self.__obs_periods}
        # Since the particles have been resampled the resulting array of logs
        # is a uniformly weighted sample of the logs, i.e., we do not need to
        # worry about different weights here.
        logs = self.__log_llhd(self.__params_original, obs_list, curr_sc,
                               period_hists)
        # check every particle is capable of producing the observation.
        inf_mask = np.where(np.logical_not(np.isfinite(logs)))
        if len(inf_mask[0]) > 0:
            msg = '{} particles have zero probability'
            raise ValueError(msg.format(len(inf_mask)))
        else:
            return logs

    def add_rows(self, hist, weights, fs_date, dates, obs_types, insert_fn):
        for date, ix, hist_ix in dates:
            if date in self._obs_dates:
                # Only compute the likelihood of the data when there is data...
                curr_obs_list = self.__obs_dict[date]
                # Get values of the log-likelihood of the current observation
                # based random samples of the hidden state from the predictive
                # density.
                logs = self._estimate_log_llhd(curr_obs_list, hist, date,
                                               hist_ix)
                # Convert the logs to *probabilities* given random samples of
                # the hidden states.
                probs = np.exp(logs)
                # Monte Carlo integration is performed to estimate the
                # *probability* of the data by averaging the probabilities
                # conditional on the hidden state for a sample of hidden states
                # drawn from the appropriate density.
                if self._resample_method == 'basic':
                    row = (date, np.mean(probs),
                           np.var(probs) /
                           self.__params_original['hist']['px_count'])
                else:
                    row = (date, np.mean(probs))
                insert_fn(row)

    def n_rows(self, start_date, end_date, n_days, n_sys, forecasting):
        """Return the number of observations in the simulation period provided
        we are forecasting.
        """
        self._forecasting = forecasting
        if forecasting:
            return 0
        else:
            num_obs = len([od for od in self._obs_dates if
                           (od >= start_date) and (od <= end_date)])
            return num_obs


class ForecastLikelihood(Table):
    """
    Estimate the likelihood of the observations under the predictive density.
    """
    def __init__(self, name='forecast_likelihood'):
        """
        Initialise a Table to hold the estimated likelihood of the observations
        under the predictive density.
        """
        Table.__init__(self, name)

    def dtype(self, params, obs_list):
        """
        Each row contains:
         - The forecast date
         - The date of the observation
         - An estimate of the likelihood of the observation
        """
        self.__params = params
        self.__steps_per_day = params['steps_per_day']
        self.__log_llhd = params['log_llhd_fn']  # **LOG**-likelihood function.
        self.__obs_list = obs_list
        self._obs_dates = [o['date'] for o in obs_list]
        self.__obs_dict = {d: [] for d in self._obs_dates}
        for o in obs_list:
            self.__obs_dict[o['date']].append(o)
        fs_date = dtype_date('fs_date')
        obs_date = dtype_date('obs_date')
        lhd_est = ('likelihood_est', np.float64)
        return [fs_date, obs_date, lhd_est]

    def _estimate_log_llhd(self, obs_list, hist, hist_ix):
        """
        Estimate the **log**-likelihood of the observations.

        TODO: Extend this to handle multiple observation streams. This is why
        the assert statement is in there.
        """
        assert len(obs_list) == 1
        obs = obs_list[0]
        prev_ix = max(0, hist_ix - obs['period'] * self.__steps_per_day)
        logs = self.__log_llhd(self.__params, [obs], hist[hist_ix, :],
                               {obs['period']: hist[prev_ix, ...]})
        return logs

    def add_rows(self, hist, weights, fs_date, dates, obs_types, insert_fn):
        """
        Add a row as defined in self.dtype for each observation we are
        forecasting.
        """
        for date, ix, hist_ix in dates:
            # Need to check there is an observation for this date and it is not
            # the observation recieved on the date we started forecasting as
            # this would be training data.
            if (date in self._obs_dates) and (date > self.__start_date):
                curr_obs_list = self.__obs_dict[date]
                logs = self._estimate_log_llhd(curr_obs_list, hist, hist_ix)
                lhd_est, _ = stats.avg_var_wt(np.exp(logs), weights[0, :])
                row = (fs_date, date, lhd_est)
                insert_fn(row)

    def n_rows(self, start_date, end_date, n_days, n_sys, forecasting):
        """
        For each forecasting run we need a row for each observation which
        occurs in the window of time we are forecasting.

        TODO: Extend this to handle multiple observation systems.
        """
        self.__forecasting = forecasting
        self.__start_date = start_date
        # We only want to add rows for observations which fall in the window of
        # time we are forecasting over.
        num_obs = len([d for d in self._obs_dates if
                       ((d > start_date) and (d <= end_date))])
        return num_obs if forecasting else 0


class PredictiveCIs(Table):
    """
    Record fixed-probability central credible intervals for the observations.

    :param probs: an array of probabilities that define the size of each
        central credible interval.
        The default value is ``numpy.uint8([0, 50, 90, 95, 99, 100])``.
    :param name: the name of the table in the output file.
    """

    def __init__(self, probs=None, name='predictive_cis'):
        Table.__init__(self, name)
        if probs is None:
            probs = np.uint8([0, 50, 90, 95, 99, 100])
        self.__probs = probs
        self.__monitor = PeakMonitor()

    def monitors(self):
        return [self.__monitor]

    def dtype(self, params, obs_list):
        self.__params = params
        self.__obs_dates = [o['date'] for o in obs_list]
        self.__obs_dict = {o['date']: o for o in obs_list}
        ulen = max(len(o['unit']) for o in obs_list)
        unit = ('unit', 'S{}'.format(ulen))
        period = ('period', np.int8)
        fs_date = dtype_date('fs_date')
        date = dtype_date()
        prob = ('prob', np.int8)
        ymin = ('ymin', np.float64)
        ymax = ('ymax', np.float64)
        return [unit, period, fs_date, date, prob, ymin, ymax]

    def n_rows(self, start_date, end_date, n_days, n_sys, forecasting):
        self.__forecasting = forecasting
        self.__sim_obs_dates = [od for od in self.__obs_dates
                                if (od > start_date and od <= end_date)]
        # Need a row for each observation, for each data source, for each prob.
        return len(self.__sim_obs_dates) * len(self.__probs) * n_sys

    def add_rows(self, hist, weights, fs_date, dates, obs_types, insert_fn):
        exp_obs = self.__monitor.expected_obs

        for (unit, period) in obs_types:
            exp_sys = exp_obs[(unit, period)]

            for date, ix, _ in dates:

                if date in self.__sim_obs_dates:
                    pred_cis = self.approximate_cis(exp_sys[ix, :],
                                                    weights[ix, :], unit)
                    date_str = str(date)
                    for cix, pctl in enumerate(self.__probs):
                        row = (unit, period, fs_date, date_str, pctl,
                               pred_cis[pctl][0], pred_cis[pctl][1])
                        insert_fn(row)

    def approximate_cis(self, mu, weights, unit):
        """Approximate the credible intervals of the predictive distribution.

        Uses a normal approximation and Newton-Raphson method to approximate
        the CIs. The method is initialised by the credible intervals of the
        mean parameter as they provide a decent starting point.
        """
        def newton_raphson(f, f_dash, x0):
            """Apply Newton-Raphson to solve f(x) == 0."""
            # Hyper parameters of the method.
            diff = 1e6
            min_diff = 0.1
            loop_count = 0
            loop_max = 10
            while (diff > min_diff) and (loop_count < loop_max):
                x1 = x0 - (f(x0) / f_dash(x0))
                diff = np.abs(x1 - x0)
                loop_count += 1
                x0 = x1
            good_exit = (not (diff > min_diff)) and (loop_count < loop_max)
            assert good_exit, "Iterations failed to converge."
            return x1

        # Initialise the observation distributions for each particle for
        # evaluation in the Newton method.
        k = self.__params['obs'][unit]['disp']
        sigma = np.sqrt(mu + (mu ** 2) / k)
        normal = norm(loc=mu, scale=sigma)

        # Generate initial guesses based on the parameter CIs.
        param_cis = stats.cred_wt(mu, weights, self.__probs)
        # Initialise the dictionary to hold the results.
        pred_cis = {p: np.zeros(2) for p in self.__probs}
        for p in self.__probs:
            # p is given as a percentage, not a probability so we need to
            # divide by 100 to get the probabilities.
            for q_ix, q in enumerate([0.5 - 0.005 * p, 0.5 + 0.005 * p]):
                x0 = param_cis[p][q_ix]
                x1 = newton_raphson(
                    lambda x: np.dot(normal.cdf(x), weights) - q,
                    lambda x: np.dot(normal.pdf(x), weights),
                    x0)
                pred_cis[p][q_ix] = x1
        return pred_cis
