from __future__ import absolute_import, division, print_function
# from __future__ import unicode_literals

import datetime
import numpy as np
import scipy.stats

import pypfilt
import pypfilt.stats as stats
from pypfilt.summary import dtype_date, Table


class VDHHS:
    """Victorian Department of Health and Human Services influenza surveillance
    system.
    """

    obs_file = 'demo-data.ssv'  # <-- This is where you put your data!
    obs_unit = 'notifications'
    obs_period = 7
    obs_source = 'VDHHS'

    @classmethod
    def obs(cls, start, end):
        """Returns the VDHHS data stream."""
        df = cls._count_df(filename=cls.obs_file)
        return [{'date': df['to'][i],
                 'value': df['count'][i],
                 'unit': cls.obs_unit,
                 'period': cls.obs_period,
                 'source': cls.obs_source}
                for i in range(df.shape[0]) if  (start <= df['from'][i]) and (df['to'][i] <= end)]

    @staticmethod
    def _count_df(filename):
        """Returns an array of the data in a notifications file."""
        def _date_col(text, fmt='%Y-%m-%d'):
            return datetime.datetime.strptime(text, fmt)

        col_types = [('year', np.int32), ('week', np.int8), ('from', '|O4'),
                     ('to', '|O4'), ('count', np.int32)]
        col_convs = {2: _date_col, 3: _date_col}
        return np.loadtxt(filename, skiprows=1, dtype=col_types,
                          converters=col_convs)

    @staticmethod
    def expect(params, op, period, pr_inf, prev, curr):
        """Determine the expected value for a given infection probability.

        :param params: The observation model parameters.
        :param op: The observation model parameters.
        :param period: The duration of the observation period (in days).
        :param pr_inf: The probability of an individual becoming infected
            during the observation period.
        """
        n = params['epifx']['popn_size']
        return op['obs_bg'] + pr_inf * op['pr_notif_inf'] * n

    @staticmethod
    def log_llhd(params, op, obs, pr_indiv, curr, hist):
        """Calculate the log-likelihood of obtaining one or more observations
        from each particle.

        :param params: The simulation parameters.
        :param op: The observation model parameters.
        :param obs: The list of observations for the current time-step.
        :param pr_indiv: A 2 x N matrix (for N particles) of the probability
            of an individual being uninfected and infected, respectively.
        :param curr: The particle state vectors.
        :param hist: The particle state histories, indexed by observation
            period.
        """
        obs_count = obs['value']
        # The (mean) background rate.
        obs_bg = op['obs_bg']
        # Observations are obtained from the entire population.
        size = params['epifx']['popn_size']
        # Notification probabilities (uninfected and infected).
        pr_notif_inf = op['pr_notif_inf']
        pr_ili = np.array([obs_bg / size, pr_notif_inf])
        # Calculate the net probability of flu notifications.
        pr_obs = np.sum(pr_ili[:, None] * pr_indiv, axis=0)
        # Calculate the expected number of failures (i.e., cases).
        mu = pr_obs * size
        # Extract the over-dispersion parameter.
        nb_k = op['disp']
        nb_pr = nb_k / (nb_k + mu)
        nbinom = scipy.stats.nbinom(nb_k, nb_pr)
        return nbinom.logpmf(obs_count)

    @classmethod
    def est_bg(cls, bg_start, bg_end):
        """Returns an estimation of the background rate based on observations
        from dates [bg_start, bg_end).

        :param bg_start: The inclusive date to start observing background.
        :param bg_end: The exclusive date to stop observing background.
        """
        df = cls._count_df(filename=cls.obs_file)
        min_ix = np.min(np.where(df['from'] >= bg_start))
        # "<=" is used because we compare to "to" not "from".
        max_ix = np.max(np.where(df['to'] <= bg_end))
        df = df[range(min_ix, max_ix + 1)]
        return np.mean(df['count'])


class DataLogLikelihood(Table):
    """
    Estimates the log-likelihood of the data during each filtering step.

    This Table estimates

        p( Y_{i+1} | Y_{1:i} )

    by Monte Carlo integration of the result of applying the LoTP.
    """

    def __init__(self, name='dataloglikelihood', seed=7):
        Table.__init__(self, name)
        self.__rng = np.random.RandomState(seed)

    def dtype(self, params, obs_list):
        """Each row contains the date of the observation, an estimate of the
        likelihood of the observation given all previous observations and an
        approximation of the standard error in this estimate.
        """
        self.__params = params
        self.__dt = 1.0 / params['steps_per_day']
        self.__sc = params['model'].state_size()
        self.__model = params['model']
        self.__log_llhd = params['log_llhd_fn']
        self.__op = params['obs']
        self.__obs_list = obs_list
        self.__obs_dates = [o['date'] for o in obs_list]
        self.__obs_periods = set(o['period'] for o in obs_list)
        self.__obs_dict = {d: [] for d in self.__obs_dates}
        for o in obs_list:
            self.__obs_dict[o['date']].append(o)

        obs_date = dtype_date('obs_date')
        log_llhd = ('loglikelihood', np.float64)
        se = ('standard_error', np.float64)
        return [obs_date, log_llhd, se]

    def n_rows(self, start_date, end_date, n_days, n_sys, forecasting):
        """Return the number of observations in the simulation period provided
        we are not forecasting.
        """
        if forecasting:
            return 0
        else:
            num_obs = len([od for od in self.__obs_dates if
                           (od >= start_date) and (od <= end_date)])
            return num_obs

    def add_rows(self, hist, weights, fs_date, dates, obs_types, insert_fn):
        """Estimate the conditional likelihood described above.

        Iterate over the dates in the current history array and if there is a
        date which has an observation estimate the likelihood of that
        observation. This is done by integrating the conditional likelihood
        with respect to

            p( X_{i+1} | Y_{1:i} )

        Prior to estimating the likelihood under the naive particles we
        re-sample to approximate a random sample so that we estimate the
        standard error of the likelihood estimate.

        Note: The state vectors are correctly populated in not "is_valid"
        particles so we do not create a mask to sieve them out.
        """
        for date, ix, hist_ix in dates:
            if date in self.__obs_dates:
                # --------------------------------------------------------------
                # 0. Alias the list of observations from this date. Since
                # epifx.obs.log_llhd() takes a list of observations from the
                # current date we want out data in this form.
                curr_obs = self.__obs_dict[date]
                # --------------------------------------------------------------
                # 1. Make a copy of the history array as we do not want to
                # change the one provided. Alias the time slice prior to the
                # current one as prev which represents the distribution just
                # prior to witnessing the current observation. Alias the current
                # time slice and set all of its values to zero as in general
                # they cannot be trusted as a re-sampling event may have
                # occurred after witnessing the current observation.
                hist_copy = np.copy(hist)
                prev = hist_copy[hist_ix - 1, ...]
                curr = hist_copy[hist_ix, ...]
                curr *= 0
                # --------------------------------------------------------------
                # 2. Re-sample using basic re-sampling to obtain an
                # (approximately) random sample from the naive density (prior to
                # re-weighting based on the current observation). This allows us
                # to estimate the standard error in the likelihood estimate.
                # This could potentially be improved upon using a variance
                # reduction technique. Care must be taken to track the indices
                # of the particles so that we can find them after conditioning
                # on the last observation. To make sure that we have the correct
                # particles we need to shuffle the earlier time slices of the
                # history.
                pseudo_params = dict(resample=dict(method='basic',
                                                   regularisation=False,
                                                   rnd=self.__rng))
                pypfilt.resample(pseudo_params, prev)
                resampled_ixs = np.copy(prev[:, -1]).astype(int)
                hist_copy[:(hist_ix - 2), ...] = hist_copy[:(hist_ix - 2),
                                                             resampled_ixs, :]
                # --------------------------------------------------------------
                # 3. Integrate the particles forward until the time of the next
                # observation. The particles must not be re-sampled after
                # re-weighting as this would destroy the estimate. The following
                # is taken from pypfilt.step(). At the end the weights are
                # copied over however as we have just re-sampled these are
                # uniform anyway.
                prev_sc = prev[..., 0:self.__sc]
                curr_sc = curr[..., 0:self.__sc]
                when = date - datetime.timedelta(days=self.__dt)
                self.__model.update(self.__params, when, self.__dt, None,
                                    prev_sc, curr_sc)
                curr[:, -2] = prev[:, -2]
                # --------------------------------------------------------------
                # 4. Evaluate the likelihood of the data for each particle. This
                # requires the state vectors at the time of the previous
                # observation, the history has already been re-ordered according
                # to the re-sampling. Unfortunately we cannot use
                # pypfilt.reweight() for this as it normalises the weights so
                # we only get the relative values. However we can steal a lot of
                # the code used in this function. If multiple observation
                # streams are in use this makes the assumption that they are
                # independent given the hidden state.
                steps_per_day = self.__params['steps_per_day']

                def hist_for(period):
                    """Return past state vectors in the appropriate order."""
                    steps_back = steps_per_day * period
                    if hist_ix - steps_back < 0:
                        raise ValueError('Attempting to access distant past.')
                    return hist_copy[hist_ix - steps_back]

                period_hists = {period: hist_for(period)
                                for period in self.__obs_periods}
                logs = self.__log_llhd(self.__params, curr_obs, curr_sc,
                                       period_hists)
                # --------------------------------------------------------------
                # 5. Book-keeping. Check that make sure that all of the
                # particles were capable of producing the observation and raise
                # an error if some were not. Since we started with a random
                # sample the standard error can be estimated simply.
                inf_mask = np.where(np.logical_not(np.isfinite(logs)))
                if len(inf_mask[0]) > 0:
                    msp = '{} particles have zero probability'
                    raise ValueError(msg.format(len(inf_mask)))
                row = (date, np.mean(logs),
                       np.var(logs) / self.__params['hist']['px_count'])
                insert_fn(row)
